#include <stdio.h>
#include <string.h>
#include <windows.h>

struct snd_fmt
{
	DWORD dwFmt;			// must be 0x11 to identify IMA ADPCM
	DWORD *pdwData;			// raw compressed sound data from the file
	DWORD dwDataSize;		// compressed data size
	DWORD dwSampleRate;		// sampling rate - usually 44100
	DWORD dwBitsPerSample;	// bits per sample - 4 for IMA ADPCM
	DWORD dwUnk1;			// unknown - probably algorithm (must be 1 for FX or 2 for music)
	DWORD dwUnk2;			// unknown - probably uncompressed data size
	DWORD dwBlockSize;		// ADPCM block size - usually 0x24 (FX) or 0x48 (music)
}s_fmt;

FILE *f1;
FILE *f2;
HMODULE hLib;
int (__stdcall *DecompressADPCM)(snd_fmt *, DWORD **, DWORD **);

void prompt()
{
	printf("Press Enter to exit...");
	fflush(stdin);
	getchar();
}

void SetSFXMode()
{
	s_fmt.dwFmt = 0x11;
	s_fmt.dwSampleRate = 44100;
	s_fmt.dwBitsPerSample = 4;
	s_fmt.dwUnk1 = 1;
	s_fmt.dwBlockSize = 0x24;
}

void SetMUSMode()
{
	s_fmt.dwFmt = 0x11;
	s_fmt.dwSampleRate = 44100;
	s_fmt.dwBitsPerSample = 4;
	s_fmt.dwUnk1 = 2;
	s_fmt.dwBlockSize = 0x48;
}

void main(int argc, char *argv[])
{
	printf("Enter The Matrix Sound Unpacker v1.0 - (c) 2003 CoDe RiPPeR\n\n");
	if ( argc < 3 )
	{
		char *exe_name = strrchr(argv[0], '\\');
		if ( !exe_name )
			exe_name = argv[0];
		else
			exe_name++;
		printf("Use: %s <sound file> <mode>\n", exe_name);
		printf("Mode can be:\n"
			   "\t-sfx\t- unpack file with default settings set for FX\n"
			   "\t-mus\t- unpack file with default settings set for music\n"
			   "\t-manual\t- manually set all required parameters\n");
		prompt();
		return;
	}
	
	hLib = LoadLibrary("mss32.dll");
	if ( !hLib )
	{
		printf("Error: File 'mss32.dll' cannot be loaded!\n");
		prompt();
		return;
	}

	DecompressADPCM = (int (__stdcall *)(snd_fmt *, DWORD **, DWORD **))
		GetProcAddress(hLib, "_AIL_decompress_ADPCM@12");
	if ( DecompressADPCM == NULL )
	{
		printf("Error: Function was not found in library 'mss32.dll'!\n");
		FreeLibrary(hLib);
		prompt();
		return;
	}

	char *mode, *s_file;
	s_file = argv[1];
	mode = argv[2];

	if ( !strcmpi(mode, "-sfx") )
	{
		SetSFXMode();
	}
	else if ( !strcmpi(mode, "-mus") )
	{
		SetMUSMode();
	}
	else if ( !strcmpi(mode, "-manual") )
	{
		s_fmt.dwFmt = 0x11;
		s_fmt.dwBitsPerSample = 4;
		printf("Set audio sampling rate: ");
		scanf("%d", &s_fmt.dwSampleRate);
		fflush(stdin);
		printf("Set ADPCM block size (hex): ");
		scanf("%x", &s_fmt.dwBlockSize);
		fflush(stdin);
		printf("Set decompression algorithm (1 or 2): ");
		scanf("%d", &s_fmt.dwUnk1);
		fflush(stdin);
	}
	else
	{
		printf("Error: Illegal parameter %s\n", mode);
		FreeLibrary(hLib);
		prompt();
		return;
	}

	char wav_file[MAX_PATH] = { 0 };
	strncpy(wav_file, s_file, strrchr(s_file, '.')-s_file);
	strcat(wav_file, ".wav");

	f1 = fopen(s_file, "rb");
	f2 = fopen(wav_file, "wb");
	if ( f1 && f2 )
	{
		DWORD dwSize;
		DWORD *pdwData;
		DWORD *pWAV;
		DWORD *pWAVSize;
				
		fseek(f1, 0, 2);
		dwSize = ftell(f1);
		rewind(f1);
		dwSize >>= 2;

		pdwData = new DWORD[dwSize];
		if ( !pdwData )
		{
			printf("Error: No free memory!\n");
			FreeLibrary(hLib);
			fcloseall();
			prompt();
			return;
		}
		if ( !fread(pdwData, dwSize, sizeof(DWORD), f1) )
		{
			printf("Error: File read error!\n");
			FreeLibrary(hLib);
			fcloseall();
			delete[] pdwData;
			prompt();
			return;
		}
		
		s_fmt.pdwData = pdwData;
		// unpacked file size is calculated by the decoder in such way:
		// s_fmt.dwDataSize (in bytes) * s_fmt.dwUnk2 * 2
		// so some fixups for music needed to reduce resulting file size
		s_fmt.dwDataSize = dwSize << 2;
		if ( s_fmt.dwUnk1 == 2 )
			s_fmt.dwUnk2 = dwSize << 2;
		else
			s_fmt.dwUnk2 = dwSize << 3;

		printf("\nDecompressing file %s...\n"
			   "Format\t\t: 0x%02X\n"
			   "Sample rate\t: %d\n"
			   "Algorithm\t: %d\n"
			   "ADPCM block size: 0x%02X\n",
			   s_file, s_fmt.dwFmt, s_fmt.dwSampleRate,
			   s_fmt.dwUnk1, s_fmt.dwBlockSize);

		if ( !DecompressADPCM(&s_fmt, &pWAV, &pWAVSize) )
		{
			printf("Error: Decompression failed!\n");
			prompt();
		}
		else
		{
			char zero[0x1000] = { 0 };
			memcpy(&dwSize, pWAV+1, sizeof(DWORD));
			// as unpacked file size was determined approximately,
			// remove zero padding (silence) at end of the file and
			// correct RIFF - WAVE header
			for (DWORD i=0; i < dwSize; i++)
			{
				if ( !memcmp((BYTE *)pWAV+i, zero, 0x20) )
				{
					if ( !memcmp((BYTE *)pWAV+i, zero, 0x1000) )
					{
						*(pWAV+1) = i;
						*(pWAV+10) = i-36;
						dwSize = i;
						break;
					}
				}
			}
			if ( !fwrite(pWAV, dwSize, 1, f2) )
			{
				printf("Error: File write error!\n");
				FreeLibrary(hLib);
				prompt();
			}
			else
			{
				printf("\nFile was successfully decompressed to %s!\n", wav_file);
//				prompt();
			}
		}
		fcloseall();
		delete[] pdwData;
	}
	else
	{
		printf("Error: Cannot open file for read/write!\n");
		prompt();
	}

	FreeLibrary(hLib);
}